@@ -78,13 +78,7 @@ CConfig::CConfig(const char* path, const Hyprlang::SConfigOptions& options_) : i
             throw "File does not exist";
     }
 
-    impl->envVariables.clear();
-    for (char** env = environ; *env; ++env) {
-        const std::string ENVVAR   = *env ? *env : "";
-        const auto        VARIABLE = ENVVAR.substr(0, ENVVAR.find_first_of('='));
-        const auto        VALUE    = ENVVAR.substr(ENVVAR.find_first_of('=') + 1);
-        impl->envVariables.push_back({VARIABLE, VALUE});
-    }
+    impl->recheckEnv();
 
     std::ranges::sort(impl->envVariables, [&](const auto& a, const auto& b) { return a.name.length() > b.name.length(); });
 
@@ -358,13 +352,31 @@ std::pair<bool, CParseResult> CConfig::configSetValueSafe(const std::string& com
                     found = true;
             }
 
+            // probably a handler
+            if (!valueName.contains(":"))
+                return {false, result};
+
             if (!found) {
                 for (auto& sc : impl->specialCategories) {
-                    if (!valueName.starts_with(sc->name))
+                    if (!valueName.starts_with(sc->name + ":"))
                         continue;
 
-                    if (!sc->isStatic && std::string{std::any_cast<const char*>(sc->values[sc->key].getValue())} != impl->currentSpecialKey)
-                        continue;
+                    if (!sc->isStatic) {
+                        const auto fieldName        = valueName.substr(sc->name.length() + 1);
+                        const auto existingKeyValue = std::string{std::any_cast<const char*>(sc->values[sc->key].getValue())};
+
+                        // When parsing the key field itself, match by the value being set.
+                        // Otherwise, match by currentSpecialKey.
+                        // This ensures multiple blocks with different key values create separate categories,
+                        // and correctly handles empty string keys.
+                        if (fieldName == sc->key) {
+                            if (existingKeyValue != value)
+                                continue;
+                        } else {
+                            if (existingKeyValue != impl->currentSpecialKey)
+                                continue;
+                        }
+                    }
 
                     VALUEIT                      = sc->values.find(valueName.substr(sc->name.length() + 1));
                     impl->currentSpecialCategory = sc.get();
@@ -528,6 +540,16 @@ CParseResult CConfig::parseVariable(const std::string& lhs, const std::string& r
     return result;
 }
 
+void CConfigImpl::recheckEnv() {
+    envVariables.clear();
+    for (char** env = environ; *env; ++env) {
+        const std::string ENVVAR   = *env ? *env : "";
+        const auto        VARIABLE = ENVVAR.substr(0, ENVVAR.find_first_of('='));
+        const auto        VALUE    = ENVVAR.substr(ENVVAR.find_first_of('=') + 1);
+        envVariables.push_back({VARIABLE, VALUE});
+    }
+}
+
 SVariable* CConfigImpl::getVariable(const std::string& name) {
     for (auto& v : envVariables) {
         if (v.name == name)
@@ -927,6 +949,10 @@ CParseResult CConfig::parse() {
     return fileParseResult;
 }
 
+void CConfig::changeRootPath(const char* path) {
+    impl->path = path;
+}
+
 CParseResult CConfig::parseRawStream(const std::string& stream) {
     CParseResult      result;
 
@@ -1030,13 +1056,13 @@ CParseResult CConfig::parseFile(const char* file) {
 }
 
 CParseResult CConfig::parseDynamic(const char* line) {
-    auto ret = parseLine(line, true);
+    auto ret                     = parseLine(line, true);
     impl->currentSpecialCategory = nullptr;
     return ret;
 }
 
 CParseResult CConfig::parseDynamic(const char* command, const char* value) {
-    auto ret = parseLine(std::string{command} + "=" + std::string{value}, true);
+    auto ret                     = parseLine(std::string{command} + "=" + std::string{value}, true);
     impl->currentSpecialCategory = nullptr;
     return ret;
 }
@@ -1044,7 +1070,8 @@ CParseResult CConfig::parseDynamic(const char* command, const char* value) {
 void CConfig::clearState() {
     impl->categories.clear();
     impl->parseError = "";
-    impl->variables  = impl->envVariables;
+    impl->recheckEnv();
+    impl->variables = impl->envVariables;
     std::erase_if(impl->specialCategories, [](const auto& e) { return !e->isStatic; });
 }
 
